package markov;

import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.stream.Collectors;

public class LookupImpl implements Lookup {


    private static final long serialVersionUID = 8382785417663574549L;
    private Lookup innerLookup = new CollectingLookup();
    private int totalCounts = 0;


    private class CollectingLookup implements Lookup {

        private static final long serialVersionUID = 8508868695745897489L;
        private final LinkedHashMap<Token, Integer> tokens = new LinkedHashMap<>();

        @Override
        public int getDistinctTokens() {
            throw newISE();
        }

        private IllegalStateException newISE() {
            return new IllegalStateException("call finishCollecting first");
        }

        @Override
        public void add(Token t) {
            Integer i = tokens.getOrDefault(t, 0);
            tokens.put(t, i + 1);
            totalCounts++;
        }

        @Override
        public Decission forRandom(double random) {
            throw newISE();
        }

        @Override
        public Decission forId(int id) {
            throw newISE();
        }

        @Override
        public Decission average() {
            throw newISE();
        }

        @Override
        public Set<Token> allPossible() {
            throw newISE();
        }

        @Override
        public void finishCollection() {
            int size = tokens.size();
            int[] finishedSums = new int[size];
            Decission[] finalData = new Decission[size];
            List<Entry<Token, Integer>> orderedEntries = tokens.entrySet()
                    .stream()
                    .sequential()
                    .sorted(Comparator.comparingInt(Entry::getValue))
                    .collect(Collectors.toList());
            int sum = 0;
            for (int i = 0; i < orderedEntries.size(); i++) {
                Entry<Token, Integer> entry = orderedEntries.get(i);
                sum += entry.getValue();
                finishedSums[i] = sum;
                finalData[i] = new Decission(entry.getKey(),
                        i,
                        LookupImpl.this,
                        (double) entry.getValue() / totalCounts);
            }
            innerLookup = new OptimizedLookup(finalData, finishedSums);
        }

        @Override
        public int getTotalCounts() {
            throw newISE();
        }
    }

    private class OptimizedLookup implements Lookup {

        private static final long serialVersionUID = -8666528642123058396L;
        private Decission[] finalData;
        private int[] finishedSums;

        OptimizedLookup(Decission[] finalData, int[] finishedSums) {
            this.finalData = finalData;
            this.finishedSums = finishedSums;
        }

        @Override
        public int getDistinctTokens() {
            return finalData.length;
        }

        @Override
        public void add(Token t) {
            throw new IllegalStateException("already optimized");
        }

        @Override
        public Decission forRandom(double random) {
            if (random < 0 || random >= 1) {
                throw new IllegalArgumentException("expected double [0; 1)");
            }
            if (totalCounts == 0) {
                return null;
            }
            int id = (int) Math.floor(random * totalCounts);
            int i = Arrays.binarySearch(finishedSums, id);
            i = Math.min(i >= 0 ? i + 1 : -1 * (i + 1), finishedSums.length - 1);
            return finalData[i];
        }

        @Override
        public Decission forId(int id) {
            if (id < 0 || id >= finalData.length || totalCounts == 0) {
                throw new IllegalArgumentException(String.format("got %d but expected id [0; %d)",
                        id,
                        finishedSums.length));
            }
            return finalData[id];
        }

        @Override
        public Decission average() {
            if (totalCounts == 0) {
                return null;
            }
            return finalData[this.finishedSums.length - 1];
        }

        @Override
        public Set<Token> allPossible() {
            return Arrays.stream(finalData).map(Decission::getToken).collect(Collectors.toSet());
        }

        @Override
        public void finishCollection() {
            throw new IllegalStateException("already finished");
        }

        @Override
        public int getTotalCounts() {
            return totalCounts;
        }

    }


    @Override
    public int getDistinctTokens() {
        return innerLookup.getDistinctTokens();
    }

    @Override
    public void add(Token t) {
        innerLookup.add(t);
    }

    @Override
    public Decission forRandom(double random) {
        return innerLookup.forRandom(random);

    }

    @Override
    public Decission forId(int id) {
        return innerLookup.forId(id);
    }

    @Override
    public Decission average() {
        return innerLookup.average();
    }

    @Override
    public Set<Token> allPossible() {
        return innerLookup.allPossible();
    }

    @Override
    public void finishCollection() {
        innerLookup.finishCollection();
    }

    @Override
    public int getTotalCounts() {
        return innerLookup.getTotalCounts();
    }


}
