package markov;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import java.util.Arrays;
import java.util.HashSet;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertIterableEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

public class LookupTests {

    LookupImpl lookup;

    @BeforeEach
    void initLookup() {
        lookup = new LookupImpl();
    }

    @Test
    void nonFinishedAddDosntThrow() {
        lookup.add(new Token("foo"));
    }

    @Test
    void emptyFinishDosntThrow() {
        lookup.finishCollection();
    }


    @Test
    void nonFinishedForRandomThrows() {
        assertThrows(IllegalStateException.class, () -> lookup.forRandom(1));
    }

    @Test
    void nonFinishedallPossibleThrows() {
        assertThrows(IllegalStateException.class, () -> lookup.allPossible());
    }

    @Test
    void nonFinishedAverageThrows() {
        assertThrows(IllegalStateException.class, () -> lookup.average());
    }

    @Test
    void nonFinishedForIdThrows() {
        assertThrows(IllegalStateException.class, () -> lookup.forId(1));
    }

    @Test
    void nonFinishedGetTotalCountsThrows() {
        assertThrows(IllegalStateException.class, () -> lookup.getTotalCounts());
    }

    @Test
    void nonFinishedgetDistinctTokensThrows() {
        assertThrows(IllegalStateException.class, () -> lookup.getDistinctTokens());
    }

    @Nested
    class FinishedEmpty {
        @BeforeEach
        void finishLookup() {
            lookup.finishCollection();
        }

        @Test
        void AddThrows() {
            assertThrows(IllegalStateException.class, () -> lookup.add(new Token("foo")));
        }

        @Test
        void finishThrows() {
            assertThrows(IllegalStateException.class, () -> lookup.finishCollection());
        }


        @Test
        void forRandom() {
            assertNull(lookup.forRandom(0));
        }

        @Test
        void allPossible() {
            assertTrue(lookup.allPossible().isEmpty());
        }

        @Test
        void average() {
            assertNull(lookup.average());
        }

        @Test
        void forId() {
            assertThrows(IllegalArgumentException.class, () -> lookup.forId(0));
            assertThrows(IllegalArgumentException.class, () -> lookup.forId(1));
            assertThrows(IllegalArgumentException.class, () -> lookup.forId(-1));
        }

        @Test
        void getTotalCounts() {
            assertEquals(0, lookup.getTotalCounts());
        }

        @Test
        void getDistinctTokens() {
            assertEquals(0, lookup.getDistinctTokens());
        }
    }

    @Nested
    class WithA {

        @BeforeEach
        void initFinishedLookupWithA() {
            lookup.add(new Token("a"));
            lookup.finishCollection();
        }

        @Test
        void forRandom() {
            assertEquals(lookup.forRandom(0).getToken(), new Token("a"));
            assertEquals(lookup.forRandom(0.5d).getToken(), new Token("a"));
            assertEquals(lookup.forRandom(0.999d).getToken(), new Token("a"));
            assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(1));
            assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(-1));
            assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(1.001d));
            assertThrows(IllegalArgumentException.class, () -> lookup.forRandom(2));
        }

        @Test
        void average() {
            assertEquals(new Token("a"), lookup.average().getToken());
        }

        @Test
        void allPossible() {
            assertIterableEquals(Arrays.asList(new Token("a")), lookup.allPossible());
        }

        @Test
        void getTotalCounts() {
            assertEquals(1, lookup.getTotalCounts());
        }

        @Test
        void getDistinctTokens() {
            assertEquals(1, lookup.getDistinctTokens());
        }

        @Test
        void forId() {
            assertEquals(lookup.forId(0).getToken(), new Token("a"));
            assertThrows(IllegalArgumentException.class, () -> lookup.forId(1));
        }

    }

    @Nested
    class WithAandB {
        @BeforeEach
        void initFinishedLookupWithAandB() {
            lookup.add(new Token("a"));
            lookup.add(new Token("b"));
            lookup.finishCollection();
        }

        @Test
        void forRandom() {

            Decission a = lookup.forRandom(0);
            Decission b = lookup.forRandom(0.9d);
            assertNotEquals(a, b);
        }

        @Test
        void average() {
            assertEquals(new Token("b"), lookup.average().getToken());
        }

        @Test
        void allPossible() {
            assertEquals(
                    new HashSet<>(Arrays.asList(new Token("a"), new Token("b"))),
                    lookup.allPossible());
        }

        @Test
        void getTotalCounts() {
            assertEquals(2, lookup.getTotalCounts());
        }

        @Test
        void getDistinctTokens() {
            assertEquals(2, lookup.getDistinctTokens());
        }

        @Test
        void forId() {
            assertEquals(lookup.forId(0).getToken(), new Token("a"));
            assertEquals(lookup.forId(1).getToken(), new Token("b"));
            assertThrows(IllegalArgumentException.class, () -> lookup.forId(2));
        }
    }

    @Nested
    class WithAandBandC {

        @BeforeEach
        void initFinishedLookupWithAandBandC() {
            lookup.add(new Token("a"));
            lookup.add(new Token("b"));
            lookup.add(new Token("c"));
            lookup.finishCollection();
        }

        @Test
        void forRandom() {
            Decission a = lookup.forRandom(0);
            Decission b = lookup.forRandom(0.51d);
            Decission c = lookup.forRandom(0.9d);
            assertEquals(new Token("a"), a.getToken());
            assertEquals(new Token("b"), b.getToken());
            assertEquals(new Token("c"), c.getToken());

        }

        @Test
        void average() {
            assertEquals(new Token("c"), lookup.average().getToken());

        }

        @Test
        void allPossible() {
            assertEquals(
                    new HashSet<>(Arrays.asList(new Token("a"), new Token("b"), new Token("c"))),
                    lookup.allPossible());
        }

        @Test
        void getTotalCounts() {
            assertEquals(3, lookup.getTotalCounts());
        }

        @Test
        void getDistinctTokens() {
            assertEquals(3, lookup.getDistinctTokens());
        }

        @Test
        void forId() {
            assertEquals(new Token("a"), lookup.forId(0).getToken());
            assertEquals(new Token("b"), lookup.forId(1).getToken());
            assertEquals(new Token("c"), lookup.forId(2).getToken());
            assertThrows(IllegalArgumentException.class, () -> lookup.forId(3));
        }
    }

    @Nested
    class WithAandBandCCC {
        @BeforeEach
        void initFinishedLookupWithAandB() {
            lookup = new LookupImpl();
            lookup.add(new Token("a"));
            lookup.add(new Token("b"));
            lookup.add(new Token("c"));
            lookup.add(new Token("c"));
            lookup.add(new Token("c"));
            lookup.finishCollection();
        }

        @Test
        void forRandom() {
            assertEquals(new Token("a"), lookup.forRandom(0).getToken());
            assertEquals(new Token("a"), lookup.forRandom(0.19d).getToken());
            assertEquals(new Token("b"), lookup.forRandom(0.2d).getToken());
            assertEquals(new Token("b"), lookup.forRandom(0.39d).getToken());
            assertEquals(new Token("c"), lookup.forRandom(0.4d).getToken());
            assertEquals(new Token("c"), lookup.forRandom(0.99d).getToken());
        }

        @Test
        void average() {
            assertEquals(new Token("c"), lookup.average().getToken());

        }

        @Test
        void allPossible() {
            assertEquals(
                    new HashSet<>(Arrays.asList(new Token("a"), new Token("b"), new Token("c"))),
                    lookup.allPossible());
        }

        @Test
        void getTotalCounts() {
            assertEquals(5, lookup.getTotalCounts());
        }

        @Test
        void getDistinctTokens() {
            assertEquals(3, lookup.getDistinctTokens());
        }

        @Test
        void forId() {
            assertEquals(new Token("a"), lookup.forId(0).getToken());
            assertEquals(new Token("b"), lookup.forId(1).getToken());
            assertEquals(new Token("c"), lookup.forId(2).getToken());
            assertThrows(IllegalArgumentException.class, () -> lookup.forId(3));
        }
    }

}
